#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re
import redis
from redis.client import Lock
from hashring import HashRing
import functools
from pipeline import Pipeline

from .helpers import format_config

_findhash = re.compile('.*\{(.*)\}.*', re.I)


class RedisShardAPI(object):

    SHARD_METHODS = set([
        "get", "set", "getset",
        "setnx", "setex",
        "incr", "decr", "exists",
        "delete", "get_type", "type", "rename",
        "expire", "ttl", "push", "persist",
        "llen", "lrange", "ltrim", "lpush", "lpop",
        "lindex", "pop", "lset",
        "lrem", "sadd", "srem", "scard",
        "sismember", "smembers",
        "zadd", "zrem", "zincrby", "zincr", "zrank",
        "zrange", "zrevrange", "zrangebyscore", "zremrangebyrank", "zrevrangebyscore",
        "zremrangebyscore", "zcard", "zscore", "zcount",
        "hget", "hset", "hdel", "hincrby", "hlen",
        "hkeys", "hvals", "hgetall", "hexists", "hmget", "hmset",
        "publish", "rpush", "rpop"
    ])

    def __init__(self, settings=None):
        self.nodes = []
        self.connections = {}
        settings = format_config(settings)
        for server in settings:
            name = server.get('name')
            conn = redis.Redis(host=server.get('host'),
                               port=server.get('port'),
                               db=server.get('db'),
                               password=server.get('password'),
                               socket_timeout=server.get('socket_timeout'),
                               unix_socket_path=server.get('unix_socket_path'),
                               )
            server['name'] = name
            if name in self.connections:
                raise ValueError("server's name config must be unique")
            self.connections[name] = conn
            self.nodes.append(name)
        self.ring = HashRing(self.nodes)

    def get_server_name(self, key):
        g = _findhash.match(key)
        if g is not None and len(g.groups()) > 0:
            key = g.groups()[0]
        name = self.ring.get_node(key)
        return name

    def get_server(self, key):
        name = self.get_server_name(key)
        return self.connections[name]

    def __wrap(self, method, *args, **kwargs):
        try:
            key = args[0]
            assert isinstance(key, basestring)
        except:
            raise ValueError("method '%s' requires a key param as the first argument" % method)
        server = self.get_server(key)
        f = getattr(server, method)
        return f(*args, **kwargs)

    def __wrap_tag(self, method, *args, **kwargs):
        key = args[0]
        if isinstance(key, basestring) and '{' in key:
            server = self.get_server(key)
        elif isinstance(key, list) and '{' in key[0]:
            server = self.get_server(key[0])
        else:
            raise ValueError("method '%s' requires tag key params as its arguments" % method)
        method = method.lstrip("tag_")
        f = getattr(server, method)
        return f(*args, **kwargs)

    def __getattr__(self, method):
        if method in self.SHARD_METHODS:
            return functools.partial(self.__wrap, method)
        elif method.startswith("tag_"):
            return functools.partial(self.__wrap_tag, method)
        else:
            raise NotImplementedError("method '%s' cannot be sharded" % method)

    #########################################
    ###  some methods implement as needed ###
    ########################################
    def brpop(self, key, timeout=0):
        if not isinstance(key, basestring):
            raise NotImplementedError("The key must be single string;mutiple keys cannot be sharded")
        server = self.get_server(key)
        return server.brpop(key, timeout)

    def blpop(self, key, timeout=0):
        if not isinstance(key, basestring):
            raise NotImplementedError("The key must be single string;mutiple keys cannot be sharded")
        server = self.get_server(key)
        return server.blpop(key, timeout)

    def keys(self, key):
        _keys = []
        for node in self.nodes:
            server = self.connections[node]
            _keys.extend(server.keys(key))
        return _keys

    def flushdb(self):
        for node in self.nodes:
            server = self.connections[node]
            server.flushdb()

    def lock(self, name, timeout=None, sleep=0.1):
        """
        Return a new Lock object using key ``name`` that mimics
        the behavior of threading.Lock.

        If specified, ``timeout`` indicates a maximum life for the lock.
        By default, it will remain locked until release() is called.

        ``sleep`` indicates the amount of time to sleep per loop iteration
        when the lock is in blocking mode and another client is currently
        holding the lock.
        """
        return Lock(self, name, timeout=timeout, sleep=sleep)

    def pipeline(self):
        return Pipeline(self)




import functools


class Pipeline(object):
    def __init__(self, shard_api):
        self.shard_api = shard_api
        self.pipelines = {}

    def get_pipeline(self, key):
        name = self.shard_api.get_server_name(key)
        if name not in self.pipelines:
            self.pipelines[name] = self.shard_api.connections[name].pipeline()
        return self.pipelines[name]

    def __wrap(self, method, *args, **kwargs):
        try:
            key = args[0]
            assert isinstance(key, basestring)
        except:
            raise ValueError("method '%s' requires a key param as the first argument" % method)
        pipeline = self.get_pipeline(key)
        f = getattr(pipeline, method)
        return f(*args, **kwargs)

    def __wrap_tag(self, method, *args, **kwargs):
        key = args[0]
        if isinstance(key, basestring) and '{' in key:
            pipeline = self.get_pipeline(key)
        elif isinstance(key, list) and '{' in key[0]:
            pipeline = self.get_pipeline(key[0])
        else:
            raise ValueError("method '%s' requires tag key params as its arguments" % method)
        method = method.lstrip("tag_")
        f = getattr(pipeline, method)
        return f(*args, **kwargs)

    def __hop_in(self, method, *args, **kwargs):
        if not isinstance(args[1], str):
            key = str(args[1])
        else:
            key = args[1]
        try:
            assert isinstance(key, basestring)
        except:
            raise ValueError("method '%s' requires a key param as the second argument" % method)
        pipeline = self.get_pipeline(key)
        if method == "hget_in":
            method = "hget"
        elif method == "hset_in":
            method = "hset"
        elif method == "hdel_in":
            method = "hdel"
        else:
            print "you can't be here"
        f = getattr(pipeline, method)
        return f(*args, **kwargs)

    def __qop_in(self, method, *args, **kwargs):
        key = "queue"
        pipeline = self.get_pipeline(key)
        if method == "rpush_in":
            method = "rpush"
        elif method == "blpop_in":
            method = "blpop"
        else:
            print "you can't be here"
        f = getattr(pipeline, method)
        return f(*args, **kwargs)

    def execute(self):
        results = []
        for name, pipeline in self.pipelines.iteritems():
            result = pipeline.execute()
            results.extend(list(result))
        return results

    def __getattr__(self, method):
        if method in [
            "get", "set", "getset",
            "setnx", "setex",
            "incr", "decr", "exists",
            "delete", "get_type", "type", "rename",
            "expire", "ttl", "push", "persist",
            "llen", "lrange", "ltrim", "lpush", "lpop",
            "lindex", "pop", "lset",
            "lrem", "sadd", "srem", "scard",
            "sismember", "smembers",
            "zadd", "zrem", "zincr", "zincrby", "zrank",
            "zrange", "zrevrange", "zrangebyscore", "zremrangebyrank", "zrevrangebyscore",
            "zremrangebyscore", "zcard", "zscore", "zcount",
            "hget", "hset", "hdel", "hincrby", "hlen",
            "hkeys", "hvals", "hgetall", "hexists", "hmget", "hmset",
            "publish", "rpush", "rpop"
        ]:
            return functools.partial(self.__wrap, method)
        elif method.startswith("tag_"):
            return functools.partial(self.__wrap_tag, method)
        elif method in ["hget_in", "hset_in", "hdel_in"]:
            return functools.partial(self.__hop_in, method)
        elif method in ["blpop_in", "rpush_in"]:
            return functools.partial(self.__qop_in, method)
        else:
            raise NotImplementedError(
                "method '%s' cannot be pipelined" % method)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup
import os


def read_file(*path):
    base_dir = os.path.dirname(__file__)
    file_path = (base_dir, ) + tuple(path)
    return file(os.path.join(*file_path)).read()

setup(
    name="redis-shard",
    url="http://blog.flyzen.com",
    license="BSD",
    author="Young King",
    author_email="yanckin@gmail.com",
    description="Redis Sharding API",
    long_description=(
        read_file("README.rst") + "\n\n" +
        "Change History\n" +
        "==============\n\n" +
        read_file("CHANGES.rst")),
    version="0.1.7",
    packages=["redis_shard"],
    include_package_data=True,
    zip_safe=False,
    install_requires=['redis', ],
    tests_require=['Nose'],
    classifiers=[
        "Programming Language :: Python",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
    ],
)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from .url import parse_url


def format_config(settings):
    """
    There's three config formats

    - list

    servers = [
        {'name':'node1','host':'127.0.0.1','port':10000,'db':0},
        {'name':'node2','host':'127.0.0.1','port':11000,'db':0},
        {'name':'node3','host':'127.0.0.1','port':12000,'db':0},
        ]

    - dict

    servers =
        { 'node1': {'host':'127.0.0.1','port':10000,'db':0},
            'node2': {'host':'127.0.0.1','port':11000,'db':0},
            'node3': {'host':'127.0.0.1','port':12000,'db':0},
        }

    - url_schema

    servers = ['redis://127.0.0.1:10000/0?name=node1',
                'redis://127.0.0.1:11000/0?name=node2',
                'redis://127.0.0.1:12000/0?name=node3'
        ]

    """
    configs = []
    if isinstance(settings, list):
        _type = type(settings[0])
        if _type == dict:
            return settings
        elif _type == str:
            for config in settings:
                configs.append(parse_url(config))
        else:
            raise ValueError("invalid server config")
    elif isinstance(settings, dict):
        for name, config in settings.items():
            config['name'] = name
            configs.append(config)
    else:
        raise ValueError("server's config must be list or dict")
    return configs
